@@ -1,6 +1,8 @@
 #include <mysql2_ext.h>
 #include <stdint.h>
 
+#include "mysql_enc_to_ruby.h"
+
 #ifdef HAVE_RUBY_ENCODING_H
 static rb_encoding *binaryEncoding;
 #endif
@@ -27,7 +29,7 @@ static rb_encoding *binaryEncoding;
  * (0*31557600) + (1*2592000) + (1*86400) + (0*3600) + (0*60) + 0
  */
 #define MYSQL2_MIN_TIME 2678400ULL
-#elif SIZEOF_INT < SIZEOF_LONG // 64bit Ruby 1.8
+#elif SIZEOF_INT < SIZEOF_LONG /* 64bit Ruby 1.8 */
 /* 0139-1-1 00:00:00 UTC
  *
  * (139*31557600) + (1*2592000) + (1*86400) + (0*3600) + (0*60) + 0
@@ -51,11 +53,9 @@ static VALUE cMysql2Result;
 static VALUE cBigDecimal, cDate, cDateTime;
 static VALUE opt_decimal_zero, opt_float_zero, opt_time_year, opt_time_month, opt_utc_offset;
 extern VALUE mMysql2, cMysql2Client, cMysql2Error;
-static VALUE intern_encoding_from_charset;
-static ID intern_new, intern_utc, intern_local, intern_encoding_from_charset_code,
-          intern_localtime, intern_local_offset, intern_civil, intern_new_offset;
+static ID intern_new, intern_utc, intern_local, intern_localtime, intern_local_offset, intern_civil, intern_new_offset;
 static VALUE sym_symbolize_keys, sym_as, sym_array, sym_database_timezone, sym_application_timezone,
-          sym_local, sym_utc, sym_cast_booleans, sym_cache_rows, sym_cast;
+          sym_local, sym_utc, sym_cast_booleans, sym_cache_rows, sym_cast, sym_stream, sym_name;
 static ID intern_merge;
 
 static void rb_mysql_result_mark(void * wrapper) {
@@ -64,22 +64,33 @@ static void rb_mysql_result_mark(void * wrapper) {
     rb_gc_mark(w->fields);
     rb_gc_mark(w->rows);
     rb_gc_mark(w->encoding);
+    rb_gc_mark(w->client);
   }
 }
 
 /* this may be called manually or during GC */
 static void rb_mysql_result_free_result(mysql2_result_wrapper * wrapper) {
   if (wrapper && wrapper->resultFreed != 1) {
+    /* FIXME: this may call flush_use_result, which can hit the socket */
     mysql_free_result(wrapper->result);
     wrapper->resultFreed = 1;
   }
 }
 
 /* this is called during GC */
-static void rb_mysql_result_free(void * wrapper) {
-  mysql2_result_wrapper * w = wrapper;
-  /* FIXME: this may call flush_use_result, which can hit the socket */
-  rb_mysql_result_free_result(w);
+static void rb_mysql_result_free(void *ptr) {
+  mysql2_result_wrapper * wrapper = ptr;
+  rb_mysql_result_free_result(wrapper);
+
+  // If the GC gets to client first it will be nil
+  if (wrapper->client != Qnil) {
+    wrapper->client_wrapper->refcount--;
+    if (wrapper->client_wrapper->refcount == 0) {
+      xfree(wrapper->client_wrapper->client);
+      xfree(wrapper->client_wrapper);
+    }
+  }
+
   xfree(wrapper);
 }
 
@@ -88,10 +99,10 @@ static void rb_mysql_result_free(void * wrapper) {
  * reliable way for us to tell this so we'll always release the GVL
  * to be safe
  */
-static VALUE nogvl_fetch_row(void *ptr) {
+static void *nogvl_fetch_row(void *ptr) {
   MYSQL_RES *result = ptr;
 
-  return (VALUE)mysql_fetch_row(result);
+  return mysql_fetch_row(result);
 }
 
 static VALUE rb_mysql_result_fetch_field(VALUE self, unsigned int idx, short int symbolize_keys) {
@@ -114,15 +125,18 @@ static VALUE rb_mysql_result_fetch_field(VALUE self, unsigned int idx, short int
 
     field = mysql_fetch_field_direct(wrapper->result, idx);
     if (symbolize_keys) {
-      VALUE colStr;
       char buf[field->name_length+1];
       memcpy(buf, field->name, field->name_length);
       buf[field->name_length] = 0;
+
+#ifdef HAVE_RB_INTERN3
+      rb_field = rb_intern3(buf, field->name_length, rb_utf8_encoding());
+      rb_field = ID2SYM(rb_field);
+#else
+      VALUE colStr;
       colStr = rb_str_new2(buf);
-#ifdef HAVE_RUBY_ENCODING_H
-      rb_enc_associate(colStr, rb_utf8_encoding());
-#endif
       rb_field = ID2SYM(rb_to_id(colStr));
+#endif
     } else {
       rb_field = rb_str_new(field->name, field->name_length);
 #ifdef HAVE_RUBY_ENCODING_H
@@ -140,20 +154,27 @@ static VALUE rb_mysql_result_fetch_field(VALUE self, unsigned int idx, short int
 
 #ifdef HAVE_RUBY_ENCODING_H
 static VALUE mysql2_set_field_string_encoding(VALUE val, MYSQL_FIELD field, rb_encoding *default_internal_enc, rb_encoding *conn_enc) {
-  // if binary flag is set, respect it's wishes
+  /* if binary flag is set, respect it's wishes */
   if (field.flags & BINARY_FLAG && field.charsetnr == 63) {
     rb_enc_associate(val, binaryEncoding);
+  } else if (!field.charsetnr) {
+    /* MySQL 4.x may not provide an encoding, binary will get the bytes through */
+    rb_enc_associate(val, binaryEncoding);
   } else {
-    // lookup the encoding configured on this field
-    VALUE new_encoding = rb_funcall(cMysql2Client, intern_encoding_from_charset_code, 1, INT2NUM(field.charsetnr));
-    if (new_encoding != Qnil) {
-      // use the field encoding we were able to match
-      rb_encoding *enc = rb_to_encoding(new_encoding);
-      rb_enc_associate(val, enc);
+    /* lookup the encoding configured on this field */
+    const char *enc_name;
+    int enc_index;
+
+    enc_name = mysql2_mysql_enc_to_rb[field.charsetnr-1];
+    if (enc_name != NULL) {
+      /* use the field encoding we were able to match */
+      enc_index = rb_enc_find_index(enc_name);
+      rb_enc_set_index(val, enc_index);
     } else {
-      // otherwise fall-back to the connection's encoding
+      /* otherwise fall-back to the connection's encoding */
       rb_enc_associate(val, conn_enc);
     }
+
     if (default_internal_enc) {
       val = rb_str_export_to_enc(val, default_internal_enc);
     }
@@ -163,11 +184,10 @@ static VALUE mysql2_set_field_string_encoding(VALUE val, MYSQL_FIELD field, rb_e
 #endif
 
 
-static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezone, int symbolizeKeys, int asArray, int castBool, int cast) {
+static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezone, int symbolizeKeys, int asArray, int castBool, int cast, MYSQL_FIELD * fields) {
   VALUE rowVal;
   mysql2_result_wrapper * wrapper;
   MYSQL_ROW row;
-  MYSQL_FIELD * fields = NULL;
   unsigned int i = 0;
   unsigned long * fieldLengths;
   void * ptr;
@@ -183,7 +203,7 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
 #endif
 
   ptr = wrapper->result;
-  row = (MYSQL_ROW)rb_thread_blocking_region(nogvl_fetch_row, ptr, RUBY_UBF_IO, 0);
+  row = (MYSQL_ROW)rb_thread_call_without_gvl(nogvl_fetch_row, ptr, RUBY_UBF_IO, 0);
   if (row == NULL) {
     return Qnil;
   }
@@ -193,7 +213,6 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
   } else {
     rowVal = rb_hash_new();
   }
-  fields = mysql_fetch_fields(wrapper->result);
   fieldLengths = mysql_fetch_lengths(wrapper->result);
   if (wrapper->fields == Qnil) {
     wrapper->numberOfFields = mysql_num_fields(wrapper->result);
@@ -217,34 +236,40 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
         }
       } else {
         switch(type) {
-        case MYSQL_TYPE_NULL:       // NULL-type field
+        case MYSQL_TYPE_NULL:       /* NULL-type field */
           val = Qnil;
           break;
-        case MYSQL_TYPE_BIT:        // BIT field (MySQL 5.0.3 and up)
-          val = rb_str_new(row[i], fieldLengths[i]);
-          break;
-        case MYSQL_TYPE_TINY:       // TINYINT field
+        case MYSQL_TYPE_BIT:        /* BIT field (MySQL 5.0.3 and up) */
           if (castBool && fields[i].length == 1) {
-            val = *row[i] == '1' ? Qtrue : Qfalse;
+            val = *row[i] == 1 ? Qtrue : Qfalse;
+          }else{
+            val = rb_str_new(row[i], fieldLengths[i]);
+          }
+          break;
+        case MYSQL_TYPE_TINY:       /* TINYINT field */
+          if (castBool && fields[i].length == 1) {
+            val = *row[i] != '0' ? Qtrue : Qfalse;
             break;
           }
-        case MYSQL_TYPE_SHORT:      // SMALLINT field
-        case MYSQL_TYPE_LONG:       // INTEGER field
-        case MYSQL_TYPE_INT24:      // MEDIUMINT field
-        case MYSQL_TYPE_LONGLONG:   // BIGINT field
-        case MYSQL_TYPE_YEAR:       // YEAR field
+        case MYSQL_TYPE_SHORT:      /* SMALLINT field */
+        case MYSQL_TYPE_LONG:       /* INTEGER field */
+        case MYSQL_TYPE_INT24:      /* MEDIUMINT field */
+        case MYSQL_TYPE_LONGLONG:   /* BIGINT field */
+        case MYSQL_TYPE_YEAR:       /* YEAR field */
           val = rb_cstr2inum(row[i], 10);
           break;
-        case MYSQL_TYPE_DECIMAL:    // DECIMAL or NUMERIC field
-        case MYSQL_TYPE_NEWDECIMAL: // Precision math DECIMAL or NUMERIC field (MySQL 5.0.3 and up)
-          if (strtod(row[i], NULL) == 0.000000){
+        case MYSQL_TYPE_DECIMAL:    /* DECIMAL or NUMERIC field */
+        case MYSQL_TYPE_NEWDECIMAL: /* Precision math DECIMAL or NUMERIC field (MySQL 5.0.3 and up) */
+          if (fields[i].decimals == 0) {
+            val = rb_cstr2inum(row[i], 10);
+          } else if (strtod(row[i], NULL) == 0.000000){
             val = rb_funcall(cBigDecimal, intern_new, 1, opt_decimal_zero);
           }else{
             val = rb_funcall(cBigDecimal, intern_new, 1, rb_str_new(row[i], fieldLengths[i]));
           }
           break;
-        case MYSQL_TYPE_FLOAT:      // FLOAT field
-        case MYSQL_TYPE_DOUBLE: {     // DOUBLE or REAL field
+        case MYSQL_TYPE_FLOAT:      /* FLOAT field */
+        case MYSQL_TYPE_DOUBLE: {     /* DOUBLE or REAL field */
           double column_to_double;
           column_to_double = strtod(row[i], NULL);
           if (column_to_double == 0.000000){
@@ -254,25 +279,35 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
           }
           break;
         }
-        case MYSQL_TYPE_TIME: {     // TIME field
-          int hour, min, sec, tokens;
-          tokens = sscanf(row[i], "%2d:%2d:%2d", &hour, &min, &sec);
-          val = rb_funcall(rb_cTime, db_timezone, 6, opt_time_year, opt_time_month, opt_time_month, INT2NUM(hour), INT2NUM(min), INT2NUM(sec));
+        case MYSQL_TYPE_TIME: {     /* TIME field */
+          int tokens;
+          unsigned int hour=0, min=0, sec=0;
+          tokens = sscanf(row[i], "%2u:%2u:%2u", &hour, &min, &sec);
+          if (tokens < 3) {
+            val = Qnil;
+            break;
+          }
+          val = rb_funcall(rb_cTime, db_timezone, 6, opt_time_year, opt_time_month, opt_time_month, UINT2NUM(hour), UINT2NUM(min), UINT2NUM(sec));
           if (!NIL_P(app_timezone)) {
             if (app_timezone == intern_local) {
               val = rb_funcall(val, intern_localtime, 0);
-            } else { // utc
+            } else { /* utc */
               val = rb_funcall(val, intern_utc, 0);
             }
           }
           break;
         }
-        case MYSQL_TYPE_TIMESTAMP:  // TIMESTAMP field
-        case MYSQL_TYPE_DATETIME: { // DATETIME field
-          unsigned int year, month, day, hour, min, sec, tokens;
+        case MYSQL_TYPE_TIMESTAMP:  /* TIMESTAMP field */
+        case MYSQL_TYPE_DATETIME: { /* DATETIME field */
+          int tokens;
+          unsigned int year=0, month=0, day=0, hour=0, min=0, sec=0, msec=0;
           uint64_t seconds;
 
-          tokens = sscanf(row[i], "%4d-%2d-%2d %2d:%2d:%2d", &year, &month, &day, &hour, &min, &sec);
+          tokens = sscanf(row[i], "%4u-%2u-%2u %2u:%2u:%2u.%6u", &year, &month, &day, &hour, &min, &sec, &msec);
+          if (tokens < 6) { /* msec might be empty */
+            val = Qnil;
+            break;
+          }
           seconds = (year*31557600ULL) + (month*2592000ULL) + (day*86400ULL) + (hour*3600ULL) + (min*60ULL) + sec;
 
           if (seconds == 0) {
@@ -282,26 +317,26 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
               rb_raise(cMysql2Error, "Invalid date: %s", row[i]);
               val = Qnil;
             } else {
-              if (seconds < MYSQL2_MIN_TIME || seconds > MYSQL2_MAX_TIME) { // use DateTime instead
+              if (seconds < MYSQL2_MIN_TIME || seconds > MYSQL2_MAX_TIME) { /* use DateTime for larger date range, does not support microseconds */
                 VALUE offset = INT2NUM(0);
                 if (db_timezone == intern_local) {
                   offset = rb_funcall(cMysql2Client, intern_local_offset, 0);
                 }
-                val = rb_funcall(cDateTime, intern_civil, 7, INT2NUM(year), INT2NUM(month), INT2NUM(day), INT2NUM(hour), INT2NUM(min), INT2NUM(sec), offset);
+                val = rb_funcall(cDateTime, intern_civil, 7, UINT2NUM(year), UINT2NUM(month), UINT2NUM(day), UINT2NUM(hour), UINT2NUM(min), UINT2NUM(sec), offset);
                 if (!NIL_P(app_timezone)) {
                   if (app_timezone == intern_local) {
                     offset = rb_funcall(cMysql2Client, intern_local_offset, 0);
                     val = rb_funcall(val, intern_new_offset, 1, offset);
-                  } else { // utc
+                  } else { /* utc */
                     val = rb_funcall(val, intern_new_offset, 1, opt_utc_offset);
                   }
                 }
-              } else {
-                val = rb_funcall(rb_cTime, db_timezone, 6, INT2NUM(year), INT2NUM(month), INT2NUM(day), INT2NUM(hour), INT2NUM(min), INT2NUM(sec));
+              } else { /* use Time, supports microseconds */
+                val = rb_funcall(rb_cTime, db_timezone, 7, UINT2NUM(year), UINT2NUM(month), UINT2NUM(day), UINT2NUM(hour), UINT2NUM(min), UINT2NUM(sec), UINT2NUM(msec));
                 if (!NIL_P(app_timezone)) {
                   if (app_timezone == intern_local) {
                     val = rb_funcall(val, intern_localtime, 0);
-                  } else { // utc
+                  } else { /* utc */
                     val = rb_funcall(val, intern_utc, 0);
                   }
                 }
@@ -310,10 +345,15 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
           }
           break;
         }
-        case MYSQL_TYPE_DATE:       // DATE field
-        case MYSQL_TYPE_NEWDATE: {  // Newer const used > 5.0
-          int year, month, day, tokens;
-          tokens = sscanf(row[i], "%4d-%2d-%2d", &year, &month, &day);
+        case MYSQL_TYPE_DATE:       /* DATE field */
+        case MYSQL_TYPE_NEWDATE: {  /* Newer const used > 5.0 */
+          int tokens;
+          unsigned int year=0, month=0, day=0;
+          tokens = sscanf(row[i], "%4u-%2u-%2u", &year, &month, &day);
+          if (tokens < 3) {
+            val = Qnil;
+            break;
+          }
           if (year+month+day == 0) {
             val = Qnil;
           } else {
@@ -321,7 +361,7 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
               rb_raise(cMysql2Error, "Invalid date: %s", row[i]);
               val = Qnil;
             } else {
-              val = rb_funcall(cDate, intern_new, 3, INT2NUM(year), INT2NUM(month), INT2NUM(day));
+              val = rb_funcall(cDate, intern_new, 3, UINT2NUM(year), UINT2NUM(month), UINT2NUM(day));
             }
           }
           break;
@@ -332,10 +372,10 @@ static VALUE rb_mysql_result_fetch_row(VALUE self, ID db_timezone, ID app_timezo
         case MYSQL_TYPE_BLOB:
         case MYSQL_TYPE_VAR_STRING:
         case MYSQL_TYPE_VARCHAR:
-        case MYSQL_TYPE_STRING:     // CHAR or BINARY field
-        case MYSQL_TYPE_SET:        // SET field
-        case MYSQL_TYPE_ENUM:       // ENUM field
-        case MYSQL_TYPE_GEOMETRY:   // Spatial fielda
+        case MYSQL_TYPE_STRING:     /* CHAR or BINARY field */
+        case MYSQL_TYPE_SET:        /* SET field */
+        case MYSQL_TYPE_ENUM:       /* ENUM field */
+        case MYSQL_TYPE_GEOMETRY:   /* Spatial fielda */
         default:
           val = rb_str_new(row[i], fieldLengths[i]);
 #ifdef HAVE_RUBY_ENCODING_H
@@ -369,6 +409,7 @@ static VALUE rb_mysql_result_fetch_fields(VALUE self) {
   GetMysql2Result(self, wrapper);
 
   defaults = rb_iv_get(self, "@query_options");
+  Check_Type(defaults, T_HASH);
   if (rb_hash_aref(defaults, sym_symbolize_keys) == Qtrue) {
     symbolizeKeys = 1;
   }
@@ -392,11 +433,13 @@ static VALUE rb_mysql_result_each(int argc, VALUE * argv, VALUE self) {
   ID db_timezone, app_timezone, dbTz, appTz;
   mysql2_result_wrapper * wrapper;
   unsigned long i;
-  int symbolizeKeys = 0, asArray = 0, castBool = 0, cacheRows = 1, cast = 1;
+  int symbolizeKeys = 0, asArray = 0, castBool = 0, cacheRows = 1, cast = 1, streaming = 0;
+  MYSQL_FIELD * fields = NULL;
 
   GetMysql2Result(self, wrapper);
 
   defaults = rb_iv_get(self, "@query_options");
+  Check_Type(defaults, T_HASH);
   if (rb_scan_args(argc, argv, "01&", &opts, &block) == 1) {
     opts = rb_funcall(defaults, intern_merge, 1, opts);
   } else {
@@ -423,6 +466,14 @@ static VALUE rb_mysql_result_each(int argc, VALUE * argv, VALUE self) {
     cast = 0;
   }
 
+  if(rb_hash_aref(opts, sym_stream) == Qtrue) {
+    streaming = 1;
+  }
+
+  if(streaming && cacheRows) {
+    rb_warn("cacheRows is ignored if streaming is true");
+  }
+
   dbTz = rb_hash_aref(opts, sym_database_timezone);
   if (dbTz == sym_local) {
     db_timezone = intern_local;
@@ -445,48 +496,81 @@ static VALUE rb_mysql_result_each(int argc, VALUE * argv, VALUE self) {
   }
 
   if (wrapper->lastRowProcessed == 0) {
-    wrapper->numberOfRows = mysql_num_rows(wrapper->result);
-    if (wrapper->numberOfRows == 0) {
+    if(streaming) {
+      /* We can't get number of rows if we're streaming, */
+      /* until we've finished fetching all rows */
+      wrapper->numberOfRows = 0;
       wrapper->rows = rb_ary_new();
-      return wrapper->rows;
+    } else {
+      wrapper->numberOfRows = mysql_num_rows(wrapper->result);
+      if (wrapper->numberOfRows == 0) {
+        wrapper->rows = rb_ary_new();
+        return wrapper->rows;
+      }
+      wrapper->rows = rb_ary_new2(wrapper->numberOfRows);
     }
-    wrapper->rows = rb_ary_new2(wrapper->numberOfRows);
   }
 
-  if (cacheRows && wrapper->lastRowProcessed == wrapper->numberOfRows) {
-    // we've already read the entire dataset from the C result into our
-    // internal array. Lets hand that over to the user since it's ready to go
-    for (i = 0; i < wrapper->numberOfRows; i++) {
-      rb_yield(rb_ary_entry(wrapper->rows, i));
+  if (streaming) {
+    if(!wrapper->streamingComplete) {
+      VALUE row;
+
+      fields = mysql_fetch_fields(wrapper->result);
+
+      do {
+        row = rb_mysql_result_fetch_row(self, db_timezone, app_timezone, symbolizeKeys, asArray, castBool, cast, fields);
+
+        if (block != Qnil && row != Qnil) {
+          rb_yield(row);
+          wrapper->lastRowProcessed++;
+        }
+      } while(row != Qnil);
+
+      rb_mysql_result_free_result(wrapper);
+
+      wrapper->numberOfRows = wrapper->lastRowProcessed;
+      wrapper->streamingComplete = 1;
+    } else {
+      rb_raise(cMysql2Error, "You have already fetched all the rows for this query and streaming is true. (to reiterate you must requery).");
     }
   } else {
-    unsigned long rowsProcessed = 0;
-    rowsProcessed = RARRAY_LEN(wrapper->rows);
-    for (i = 0; i < wrapper->numberOfRows; i++) {
-      VALUE row;
-      if (cacheRows && i < rowsProcessed) {
-        row = rb_ary_entry(wrapper->rows, i);
-      } else {
-        row = rb_mysql_result_fetch_row(self, db_timezone, app_timezone, symbolizeKeys, asArray, castBool, cast);
-        if (cacheRows) {
-          rb_ary_store(wrapper->rows, i, row);
+    if (cacheRows && wrapper->lastRowProcessed == wrapper->numberOfRows) {
+      /* we've already read the entire dataset from the C result into our */
+      /* internal array. Lets hand that over to the user since it's ready to go */
+      for (i = 0; i < wrapper->numberOfRows; i++) {
+        rb_yield(rb_ary_entry(wrapper->rows, i));
+      }
+    } else {
+      unsigned long rowsProcessed = 0;
+      rowsProcessed = RARRAY_LEN(wrapper->rows);
+      fields = mysql_fetch_fields(wrapper->result);
+
+      for (i = 0; i < wrapper->numberOfRows; i++) {
+        VALUE row;
+        if (cacheRows && i < rowsProcessed) {
+          row = rb_ary_entry(wrapper->rows, i);
+        } else {
+          row = rb_mysql_result_fetch_row(self, db_timezone, app_timezone, symbolizeKeys, asArray, castBool, cast, fields);
+          if (cacheRows) {
+            rb_ary_store(wrapper->rows, i, row);
+          }
+          wrapper->lastRowProcessed++;
         }
-        wrapper->lastRowProcessed++;
-      }
 
-      if (row == Qnil) {
-        // we don't need the mysql C dataset around anymore, peace it
+        if (row == Qnil) {
+          /* we don't need the mysql C dataset around anymore, peace it */
+          rb_mysql_result_free_result(wrapper);
+          return Qnil;
+        }
+
+        if (block != Qnil) {
+          rb_yield(row);
+        }
+      }
+      if (wrapper->lastRowProcessed == wrapper->numberOfRows) {
+        /* we don't need the mysql C dataset around anymore, peace it */
         rb_mysql_result_free_result(wrapper);
-        return Qnil;
       }
-
-      if (block != Qnil) {
-        rb_yield(row);
-      }
-    }
-    if (wrapper->lastRowProcessed == wrapper->numberOfRows) {
-      // we don't need the mysql C dataset around anymore, peace it
-      rb_mysql_result_free_result(wrapper);
     }
   }
 
@@ -497,12 +581,19 @@ static VALUE rb_mysql_result_count(VALUE self) {
   mysql2_result_wrapper *wrapper;
 
   GetMysql2Result(self, wrapper);
-
-  return INT2FIX(mysql_num_rows(wrapper->result));
+  if(wrapper->resultFreed) {
+    if (wrapper->streamingComplete){
+      return LONG2NUM(wrapper->numberOfRows);
+    } else {
+      return LONG2NUM(RARRAY_LEN(wrapper->rows));
+    }
+  } else {
+    return INT2FIX(mysql_num_rows(wrapper->result));
+  }
 }
 
 /* Mysql2::Result */
-VALUE rb_mysql_result_to_obj(MYSQL_RES * r) {
+VALUE rb_mysql_result_to_obj(VALUE client, VALUE encoding, VALUE options, MYSQL_RES *r) {
   VALUE obj;
   mysql2_result_wrapper * wrapper;
   obj = Data_Make_Struct(cMysql2Result, mysql2_result_wrapper, rb_mysql_result_mark, rb_mysql_result_free, wrapper);
@@ -513,8 +604,16 @@ VALUE rb_mysql_result_to_obj(MYSQL_RES * r) {
   wrapper->result = r;
   wrapper->fields = Qnil;
   wrapper->rows = Qnil;
-  wrapper->encoding = Qnil;
+  wrapper->encoding = encoding;
+  wrapper->streamingComplete = 0;
+  wrapper->client = client;
+  wrapper->client_wrapper = DATA_PTR(client);
+  wrapper->client_wrapper->refcount++;
+
   rb_obj_call_init(obj, 0, NULL);
+
+  rb_iv_set(obj, "@query_options", options);
+
   return obj;
 }
 
@@ -529,9 +628,6 @@ void init_mysql2_result() {
   rb_define_method(cMysql2Result, "count", rb_mysql_result_count, 0);
   rb_define_alias(cMysql2Result, "size", "count");
 
-  intern_encoding_from_charset = rb_intern("encoding_from_charset");
-  intern_encoding_from_charset_code = rb_intern("encoding_from_charset_code");
-
   intern_new          = rb_intern("new");
   intern_utc          = rb_intern("utc");
   intern_local        = rb_intern("local");
@@ -551,9 +647,11 @@ void init_mysql2_result() {
   sym_application_timezone  = ID2SYM(rb_intern("application_timezone"));
   sym_cache_rows     = ID2SYM(rb_intern("cache_rows"));
   sym_cast           = ID2SYM(rb_intern("cast"));
+  sym_stream         = ID2SYM(rb_intern("stream"));
+  sym_name           = ID2SYM(rb_intern("name"));
 
   opt_decimal_zero = rb_str_new2("0.0");
-  rb_global_variable(&opt_decimal_zero); //never GC
+  rb_global_variable(&opt_decimal_zero); /*never GC */
   opt_float_zero = rb_float_new((double)0);
   rb_global_variable(&opt_float_zero);
   opt_time_year = INT2NUM(2000);
