@@ -3,11 +3,36 @@ require 'spec_helper'
 
 describe Mysql2::Result do
   before(:each) do
-    @client = Mysql2::Client.new :host => "localhost", :username => "root", :database => 'test'
+    @result = @client.query "SELECT 1"
   end
 
-  before(:each) do
-    @result = @client.query "SELECT 1"
+  it "should maintain a count while streaming" do
+    result = @client.query('SELECT 1')
+
+    result.count.should eql(1)
+    result.each.to_a
+    result.count.should eql(1)
+  end
+
+  it "should set the actual count of rows after streaming" do
+      @client.query "USE test"
+      result = @client.query("SELECT * FROM mysql2_test", :stream => true, :cache_rows => false)
+      result.count.should eql(0)
+      result.each {|r|  }
+      result.count.should eql(1)
+  end
+
+  it "should not yield nil at the end of streaming" do
+    result = @client.query('SELECT * FROM mysql2_test', :stream => true)
+    result.each { |r| r.should_not be_nil}
+  end
+
+  it "#count should be zero for rows after streaming when there were no results " do
+      @client.query "USE test"
+      result = @client.query("SELECT * FROM mysql2_test WHERE null_test IS NOT NULL", :stream => true, :cache_rows => false)
+      result.count.should eql(0)
+      result.each.to_a
+      result.count.should eql(0)
   end
 
   it "should have included Enumerable" do
@@ -73,6 +98,25 @@ describe Mysql2::Result do
       result = @client.query "SELECT 1", :cache_rows => false
       result.first.object_id.should_not eql(result.first.object_id)
     end
+
+    it "should yield different value for #first if streaming" do
+      result = @client.query "SELECT 1 UNION SELECT 2", :stream => true, :cache_rows => false
+      result.first.should_not eql(result.first)
+    end
+
+    it "should yield the same value for #first if streaming is disabled" do
+      result = @client.query "SELECT 1 UNION SELECT 2", :stream => false
+      result.first.should eql(result.first)
+    end
+
+    it "should throw an exception if we try to iterate twice when streaming is enabled" do
+      result = @client.query "SELECT 1 UNION SELECT 2", :stream => true, :cache_rows => false
+
+      expect {
+        result.each.to_a
+        result.each.to_a
+      }.to raise_exception(Mysql2::Error)
+    end
   end
 
   context "#fields" do
@@ -100,11 +144,11 @@ describe Mysql2::Result do
     it "should return nil values for NULL and strings for everything else when :cast is false" do
       result = @client.query('SELECT null_test, tiny_int_test, bool_cast_test, int_test, date_test, enum_test FROM mysql2_test WHERE bool_cast_test = 1 LIMIT 1', :cast => false).first
       result["null_test"].should be_nil
-      result["tiny_int_test"].should  == "1"
-      result["bool_cast_test"].should == "1"
-      result["int_test"].should       == "10"
-      result["date_test"].should      == "2010-04-04"
-      result["enum_test"].should      == "val1"
+      result["tiny_int_test"].should  eql("1")
+      result["bool_cast_test"].should eql("1")
+      result["int_test"].should       eql("10")
+      result["date_test"].should      eql("2010-04-04")
+      result["enum_test"].should      eql("val1")
     end
 
     it "should return nil for a NULL value" do
@@ -112,11 +156,16 @@ describe Mysql2::Result do
       @test_result['null_test'].should eql(nil)
     end
 
-    it "should return Fixnum for a BIT value" do
+    it "should return String for a BIT(64) value" do
       @test_result['bit_test'].class.should eql(String)
       @test_result['bit_test'].should eql("\000\000\000\000\000\000\000\005")
     end
 
+    it "should return String for a BIT(1) value" do
+      @test_result['single_bit_test'].class.should eql(String)
+      @test_result['single_bit_test'].should eql("\001")
+    end
+
     it "should return Fixnum for a TINYINT value" do
       [Fixnum, Bignum].should include(@test_result['tiny_int_test'].class)
       @test_result['tiny_int_test'].should eql(1)
@@ -127,11 +176,29 @@ describe Mysql2::Result do
       id1 = @client.last_id
       @client.query 'INSERT INTO mysql2_test (bool_cast_test) VALUES (0)'
       id2 = @client.last_id
+      @client.query 'INSERT INTO mysql2_test (bool_cast_test) VALUES (-1)'
+      id3 = @client.last_id
 
       result1 = @client.query 'SELECT bool_cast_test FROM mysql2_test WHERE bool_cast_test = 1 LIMIT 1', :cast_booleans => true
       result2 = @client.query 'SELECT bool_cast_test FROM mysql2_test WHERE bool_cast_test = 0 LIMIT 1', :cast_booleans => true
+      result3 = @client.query 'SELECT bool_cast_test FROM mysql2_test WHERE bool_cast_test = -1 LIMIT 1', :cast_booleans => true
       result1.first['bool_cast_test'].should be_true
       result2.first['bool_cast_test'].should be_false
+      result3.first['bool_cast_test'].should be_true
+
+      @client.query "DELETE from mysql2_test WHERE id IN(#{id1},#{id2},#{id3})"
+    end
+
+    it "should return TrueClass or FalseClass for a BIT(1) value if :cast_booleans is enabled" do
+      @client.query 'INSERT INTO mysql2_test (single_bit_test) VALUES (1)'
+      id1 = @client.last_id
+      @client.query 'INSERT INTO mysql2_test (single_bit_test) VALUES (0)'
+      id2 = @client.last_id
+
+      result1 = @client.query "SELECT single_bit_test FROM mysql2_test WHERE id = #{id1}", :cast_booleans => true
+      result2 = @client.query "SELECT single_bit_test FROM mysql2_test WHERE id = #{id2}", :cast_booleans => true
+      result1.first['single_bit_test'].should be_true
+      result2.first['single_bit_test'].should be_false
 
       @client.query "DELETE from mysql2_test WHERE id IN(#{id1},#{id2})"
     end
@@ -182,7 +249,7 @@ describe Mysql2::Result do
     end
 
     if 1.size == 4 # 32bit
-      if RUBY_VERSION =~ /1.9/
+      unless RUBY_VERSION =~ /1.8/
         klass = Time
       else
         klass = DateTime
@@ -200,7 +267,7 @@ describe Mysql2::Result do
         r.first['test'].class.should eql(klass)
       end
     elsif 1.size == 8 # 64bit
-      if RUBY_VERSION =~ /1.9/
+      unless RUBY_VERSION =~ /1.8/
         it "should return Time when timestamp is < 1901-12-13 20:45:52" do
           r = @client.query("SELECT CAST('1901-12-13 20:45:51' AS DATETIME) as test")
           r.first['test'].class.should eql(Time)
@@ -255,10 +322,11 @@ describe Mysql2::Result do
           result = @client.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
           result['enum_test'].encoding.should eql(Encoding.find('utf-8'))
 
-          client2 = Mysql2::Client.new :encoding => 'ascii'
+          client2 = Mysql2::Client.new(DatabaseCredentials['root'].merge(:encoding => 'ascii'))
           client2.query "USE test"
           result = client2.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
           result['enum_test'].encoding.should eql(Encoding.find('us-ascii'))
+          client2.close
         end
 
         it "should use Encoding.default_internal" do
@@ -284,10 +352,11 @@ describe Mysql2::Result do
           result = @client.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
           result['set_test'].encoding.should eql(Encoding.find('utf-8'))
 
-          client2 = Mysql2::Client.new :encoding => 'ascii'
+          client2 = Mysql2::Client.new(DatabaseCredentials['root'].merge(:encoding => 'ascii'))
           client2.query "USE test"
           result = client2.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
           result['set_test'].encoding.should eql(Encoding.find('us-ascii'))
+          client2.close
         end
 
         it "should use Encoding.default_internal" do
@@ -366,10 +435,11 @@ describe Mysql2::Result do
               result = @client.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
               result[field].encoding.should eql(Encoding.find('utf-8'))
 
-              client2 = Mysql2::Client.new :encoding => 'ascii'
+              client2 = Mysql2::Client.new(DatabaseCredentials['root'].merge(:encoding => 'ascii'))
               client2.query "USE test"
               result = client2.query("SELECT * FROM mysql2_test ORDER BY id DESC LIMIT 1").first
               result[field].encoding.should eql(Encoding.find('us-ascii'))
+              client2.close
             end
 
             it "should use Encoding.default_internal" do
